#include "matrix.hh"
#include "lll_uses.hh"
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <dsdp/dsdp5.h>

const bool reinject_solution = true, reduce_for_numeric = false;

enum tri_state { TRI_TRUE, TRI_FALSE, TRI_NOT_YET };

class sdp_problem {
  ff_vector costs, y_cur, sol_f;
  qq_vector yq_cur;
  std::vector<unsigned> block_dims;
  std::vector<unsigned> offsets;
  std::vector<qq_matrix> positive_parts;
  std::vector<bool> is_sdp_block;
  generator_system gens;

  unsigned where(unsigned block, unsigned i, unsigned j) {
    if (i > j) {
      unsigned k=i;
      i=j;
      j=k;
    }
    return offsets[block] + (j*(j+1)/2) + i;
  }

  qq_matrix original_matrix(const qq_vector& v, unsigned block);
  qq_matrix reduced_matrix(const qq_vector& v, unsigned block);

  ff_matrix original_matrix_f(const ff_vector& v, unsigned block);
  ff_matrix reduced_matrix_f(const ff_vector& v, unsigned block);

public:
  sdp_problem(unsigned space_dimension,
	      const std::vector<unsigned>& block_dims);

  static sdp_problem load_from_dat_q_sparse(std::istream& in);
  static sdp_problem load_from_dat_q_sparse(const char *filename);

  void set_cost(unsigned i, double cost) {
    costs[i] = cost;
  }

  void set_origin_coeff(unsigned block, unsigned i, unsigned j, const QQ& value) {
    assert(block < block_dims.size());
    assert(i <= j);
    assert(j < block_dims[block]);
    gens.set_origin_coeff(where(block, i, j), value);
  }

  void set_vector_coeff(unsigned index, unsigned block, unsigned i, unsigned j, const QQ& value) {
    assert(block < block_dims.size());
    assert(i <= j);
    assert(j < block_dims[block]);
    gens.set_vector_coeff(index, where(block, i, j), value);
  }

  QQ get_origin_coeff(unsigned block, unsigned i, unsigned j) {
    return gens.get_origin_coeff(where(block, i, j));
  }

  QQ get_vector_coeff(unsigned index, unsigned block, unsigned i, unsigned j) {
    return gens.get_vector_coeff(index, where(block, i, j));
  }

  void reduce_to_positive();
  void reduce_to_positive_block(unsigned block);
  void numeric_solve();

  bool check_solution(const qq_vector& yq);
  tri_state check_and_reduce();
  bool solve();

  qq_vector get_y() {
    return yq_cur;
  }
};

inline static unsigned total_size(const std::vector<unsigned>& block_dims) {
  unsigned total = 0;
  for(unsigned i=0; i<block_dims.size(); i++) {
    total += block_dims[i]*(block_dims[i]+1)/2;
  }
  return total;
}

sdp_problem::sdp_problem(unsigned space_dimension,
			 const std::vector<unsigned>& block_dims_p) :
  costs(space_dimension),
  block_dims(block_dims_p),
  offsets(block_dims.size()+1),
  positive_parts(block_dims.size(), qq_matrix(0, 0)),
  is_sdp_block(block_dims.size()),
  gens(space_dimension, total_size(block_dims)) {

  unsigned total=0;
  for(unsigned block=0; block<block_dims.size(); block++) {
    offsets[block] = total;
    total += block_dims[block]*(block_dims[block]+1)/2;
  }
  offsets[block_dims.size()] = total;
}

sdp_problem sdp_problem::load_from_dat_q_sparse(std::istream& in) {
  unsigned space_dimension, n_blocks;
  in >> space_dimension;
  in >> n_blocks;
  std::vector<unsigned> block_dims(n_blocks);
  for(unsigned i=0; i<n_blocks; i++) {
    in >> block_dims[i];
  }
  sdp_problem problem(space_dimension, block_dims);
  for(unsigned i=0; i<space_dimension; i++) {
    double cost;
    in >> cost;
    problem.set_cost(i, cost);
  }
  while (! in.eof()) {
    unsigned index, block, i, j;
    QQ value;
    in >> index;
    in >> block;
    in >> i;
    in >> j;
    in >> value;
    if (index == 0) {
      problem.set_origin_coeff(block-1, i-1, j-1, -value);
    } else {
      problem.set_vector_coeff(index-1, block-1, i-1, j-1, value);
    }
  }
  return problem;
}

sdp_problem sdp_problem::load_from_dat_q_sparse(const char *filename) {
  std::ifstream in(filename);
  sdp_problem problem = load_from_dat_q_sparse(in);
  in.close();
  return problem;
}

void sdp_problem::reduce_to_positive_block(unsigned block) {
  unsigned dim= block_dims[block];
  qq_matrix stacked(dim*(costs.size()+1), dim);

  unsigned total_rows=0;
  for(unsigned i=0; i<dim; i++) {
    for(unsigned j=0; j<dim; j++) {
      QQ c = get_origin_coeff(block, i, j);
      if (c.isZero()) continue;
      stacked(total_rows+i, j) = c;
    }
  }
  total_rows += dim;
  for(unsigned index=0; index<costs.size(); index++) {
    for(unsigned i=0; i<dim; i++) {
      for(unsigned j=0; j<dim; j++) {
	QQ c = get_vector_coeff(index, block, i, j);
	if (c.isZero()) continue;
	stacked(total_rows+i, j) = c;
      }
    }
    total_rows += dim;
  }
  assert(total_rows = stacked.size1());

  unsigned rank = echelonize(stacked);
  positive_parts[block] = subrange(stacked, 0, rank, 0, stacked.size2());
  std::cout << "block " << block << " has rank " << rank << std::endl;
}

void sdp_problem::reduce_to_positive() {
  for(unsigned block=0; block < block_dims.size(); block++) {
    reduce_to_positive_block(block);
  }
}

class numerical_sdp_solver {
  DSDP dsdp;
  SDPCone sdpcone;
  BCone bcone;
  std::vector<double *> arrays_to_be_deleted;
  int info;
  unsigned nr_variables;

public:
  int get_info() {
    return info;
  }

  void sdp_cone_view() {
    info = SDPConeView2(sdpcone);
  }

  void view() {
    info = DSDPView(dsdp);
  }

  numerical_sdp_solver(const vector<double>& costs,
		       const std::vector<unsigned>& block_sizes);

  ~numerical_sdp_solver();

  void set_matrix(unsigned var, unsigned block, const qq_matrix& mat);
  void solve();
  ff_vector get_y();
  void set_y0(const ff_vector& y);
  void set_r0(double r0) {
    info = DSDPSetR0(dsdp, r0);
  }
};

numerical_sdp_solver::numerical_sdp_solver(
    const vector<double>& costs,
    const std::vector<unsigned>& block_sizes):
  info(0), nr_variables(costs.size()) {
  info = DSDPCreate(costs.size(),&dsdp);
  for(unsigned i=0; i<costs.size(); i++) {
    //std::cout << "cost = " << costs[i] << std::endl;
    info = DSDPSetDualObjective(dsdp, i+1, costs[i]);
    // info = DSDPSetY0(dsdp, i+1, costs[i]>0.0 ? -0.0 : +0.0);
  }

  info = DSDPCreateSDPCone(dsdp, block_sizes.size(), &sdpcone);
  for(unsigned i=0; i<block_sizes.size(); i++) {
    info = SDPConeSetBlockSize(sdpcone, i, block_sizes[i]);
  }
  info = DSDPReuseMatrix(dsdp, 10);
  DSDPSetStandardMonitor(dsdp, 1);
}

numerical_sdp_solver::~numerical_sdp_solver() {
  DSDPDestroy(dsdp);
  for(unsigned i=0; i<arrays_to_be_deleted.size(); i++) {
    delete[] arrays_to_be_deleted[i];
  }
}

void numerical_sdp_solver::set_matrix(unsigned var,
				     unsigned block,
				     const qq_matrix& mat) {
  assert(mat.size1() == mat.size2());
  unsigned dim=mat.size1();
  double *values = new double[dim*(dim+1)/2];
  arrays_to_be_deleted.push_back(values);
  unsigned pos = 0;
  for(unsigned i=0; i<dim; i++) {
    for(unsigned j=0; j<=i; j++) {
      assert(mat(i,j) == mat(j,i));
      values[pos++] = mat(i,j).get_d();
    }
  }
  assert(pos == dim*(dim+1)/2);
  info = SDPConeSetADenseVecMat(sdpcone, block, var, dim, 1.0, values, pos);
}

void numerical_sdp_solver::solve() {
  info = DSDPSetup(dsdp);
  info = DSDPSolve(dsdp);
}

ff_vector numerical_sdp_solver::get_y() {
  double y[nr_variables];
  info = DSDPGetY(dsdp, y, nr_variables);
  vector<double> ret(nr_variables); // DM maybe not needed
  for(unsigned i=0; i<nr_variables; i++) {
    ret[i] = y[i];
  }
  return ret;
}

void numerical_sdp_solver::set_y0(const ff_vector& y0) {
  for(unsigned i=0; i<y0.size(); i++) {
    DSDPSetY0(dsdp, i+1, y0[i]);
  }
}

/*
qq_matrix sdp_problem::original_matrix(const qq_vector& v, unsigned block) {
  unsigned dim=block_dims[block];
  qq_matrix ret(dim, dim);
  for(unsigned i=0; i<dim; i++) {
    for(unsigned j=0; j<dim; j++) {
      ret(i, j) = v[where(block, i, j)];
    }
  }
  return ret;
}
*/

qq_matrix sdp_problem::original_matrix(const qq_vector& v, unsigned block) {
  unsigned base = offsets[block], dim = block_dims[block];
  qq_matrix m(dim, dim);
  for(unsigned i=0; i<dim; i++) {
    vector_range<const qq_vector> vr(v, range(base, base+i+1));
    matrix_vector_slice<qq_matrix>(m, slice(0, 1, i+1), slice(i, 0, i+1)) = vr;
    matrix_vector_slice<qq_matrix>(m, slice(i, 0, i+1), slice(0, 1, i+1)) = vr;
    base += i+1;
  }
  return m;
}

qq_matrix sdp_problem::reduced_matrix(const qq_vector& v, unsigned block) {
  if (positive_parts[block].size1() == positive_parts[block].size2()) {
    // full rank: no need to reduce the matrix!
    return original_matrix(v, block);
  }
  qq_matrix tmp = prod(original_matrix(v, block),
		       trans(positive_parts[block]));
  return prod(positive_parts[block], tmp);
}

ff_matrix sdp_problem::original_matrix_f(const ff_vector& v, unsigned block) {
  unsigned base = offsets[block], dim = block_dims[block];
  ff_matrix m(dim, dim);
  for(unsigned i=0; i<dim; i++) {
    vector_range<const ff_vector> vr(v, range(base, base+i+1));
    matrix_vector_slice<ff_matrix>(m, slice(0, 1, i+1), slice(i, 0, i+1)) = vr;
    matrix_vector_slice<ff_matrix>(m, slice(i, 0, i+1), slice(0, 1, i+1)) = vr;
    base += i+1;
  }
  return m;
}

ff_matrix sdp_problem::reduced_matrix_f(const ff_vector& v, unsigned block) {
  if (positive_parts[block].size1() == positive_parts[block].size2()) {
    // full rank: no need to reduce the matrix!
    return original_matrix_f(v, block);
  }
  ff_matrix pos = qq_to_ff_matrix(positive_parts[block]);
  ff_matrix tmp = prod(original_matrix_f(v, block), trans(pos));
  return prod(pos, tmp);
}

bool sdp_problem::check_solution(const qq_vector& yq) {
  bool all_sdp = true;
  qq_vector sol = gens.at(yq);
  for(unsigned block=0; block<block_dims.size(); block++) {
    all_sdp &= (is_sdp_block[block] = is_sdp(reduced_matrix(sol, block)));
  }
  return all_sdp;
}

void sdp_problem::numeric_solve() {
  std::vector<unsigned> new_block_sizes(block_dims.size());
  if (reduce_for_numeric) {
    for(unsigned i=0; i<block_dims.size(); i++) {
      new_block_sizes[i] = positive_parts[i].size1();
    }
  } else {
    for(unsigned i=0; i<block_dims.size(); i++) {
      new_block_sizes[i] = block_dims[i];
    }   
  }
  numerical_sdp_solver solver(costs, new_block_sizes);

  if (reduce_for_numeric) {
    for(unsigned block=0; block<block_dims.size(); block++) {
      qq_matrix mat = reduced_matrix(gens.get_f0(), block);
      std::cout << "block = " << block << std::endl;
      solver.set_matrix(0, block, mat);
      for(unsigned var=0; var<gens.dimension(); var++) {
	std::cout << "var = " << var << std::endl;
	mat = reduced_matrix(-gens.get_fi(var), block);
	solver.set_matrix(var+1, block, mat);
      }
    }
  } else {
    for(unsigned block=0; block<block_dims.size(); block++) {
      qq_matrix mat = original_matrix(gens.get_f0(), block);
      std::cout << "block = " << block << std::endl;
      solver.set_matrix(0, block, mat);
      for(unsigned var=0; var<gens.dimension(); var++) {
	std::cout << "var = " << var << std::endl;
	mat = original_matrix(-gens.get_fi(var), block);
	solver.set_matrix(var+1, block, mat);
      }
    }
  }

  std::cout << "proposed y: " << y_cur << std::endl;
  ff_vector y_old = y_cur;
  if (reinject_solution && y_cur.size() > 0) {
    solver.set_y0(y_cur);
    solver.set_r0(1e-6);
    std::cout << "reinjecting info" << solver.get_info() << std::endl;
  }
  //solver.sdp_cone_view();
  std::cout << "begin sdp solving" << std::endl;
  solver.solve();
  std::cout << "solver info = " << solver.get_info() << std::endl; 
  //solver.view();

  std::cout << "end sdp solving" << std::endl;
  y_cur = solver.get_y();
  if (reinject_solution) {
    std::cout << "obtained y: " << y_cur << std::endl;
    if (y_old.size() > 0) {
      std::cout << "move distance: " << norm_2(y_cur-y_old) << std::endl;
    }
  }
}

tri_state sdp_problem::check_and_reduce() {
  //std::cout << "info = " << solver.get_info() << std::endl;

  yq_cur.resize(y_cur.size(), false);
  for(unsigned i=0; i<yq_cur.size(); i++) {
    yq_cur[i] = rationalize(y_cur[i]);
    //std::cout << "y[" << (i+1) << "] = " << y_cur[i] << " ~ " << yq_cur[i] << std::endl;
  }
  if (check_solution(yq_cur)) return TRI_TRUE;

  sol_f = gens.at(y_cur);
 
  unsigned nr_constraints = 0;
  for(unsigned block=0; block<block_dims.size(); block++) {
    if (is_sdp_block[block]) continue;
    nr_constraints += block_dims[block];
  }
  assert(nr_constraints > 0);
  qq_matrix constraints(nr_constraints, gens.ambient_dimension()+1);
  constraints.clear();

  std::cout << "computing additional constraints" << std::endl;
  for(unsigned block=0, ibase=0; block<block_dims.size(); block++) {
    if (is_sdp_block[block]) continue;

    ff_matrix mf = reduced_matrix_f(sol_f, block);
    qq_vector kvec = prod(left_near_kernel_vector(mf), positive_parts[block]);
    assert(kvec.size() == block_dims[block]);

    for(qq_vector::const_iterator pos=kvec.begin();
	pos != kvec.end(); pos++) {
      if ((*pos).isZero()) continue;
      QQ coeff = *pos;
      for(unsigned i=0; i<block_dims[block]; i++) {
	constraints(i+ibase, where(block, i, pos.index())) = coeff;
      }
    }
    ibase += block_dims[block];
  }

  std::cout << "constraining system" << std::endl;

  gens.constrain(constraints);
  if (gens.is_empty()) return TRI_FALSE;

  costs.resize(gens.dimension(), false);
  costs.clear();

  if (reinject_solution) {
    y_cur = gens.project(sol_f);
  }

  yq_cur.resize(gens.dimension(), false);

  return TRI_NOT_YET;
}

bool sdp_problem::solve() {
  while (true) {
    std::cout << "reducing to positive" << std::endl;
    reduce_to_positive();
    std::cout << "numeric solving" << std::endl;
    numeric_solve();
    tri_state v = check_and_reduce();
    if (v == TRI_FALSE) return false;
    if (v == TRI_TRUE) return true;
    for(unsigned count=0; count<4; count++) {
      reduce_to_positive();
      tri_state v = check_and_reduce();
      if (v == TRI_FALSE) return false;
      if (v == TRI_TRUE) return true;
    }
  }
}

int main(int argc, char **argv) {
  if (argc != 2) exit(1);
  std::cout << "loading" << std::endl;
  sdp_problem problem=sdp_problem::load_from_dat_q_sparse(argv[1]);
  bool answer = problem.solve();
  std::cout << "Final answer: " << answer << std::endl;
}

